/**
 * File for general "Load Balancing" logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "load_balance_general.h"
    #include "structs.h"
    #include "terminal_commands.h"
#endif


/**
 * Displays provided worker data through master process.
 */
void main_display_status(thread_struct* thread_args_ptr, int process_num, int process_load_status, int init_run) {
    int index;
    char* log_msg = calloc(256, sizeof(char));

    log("");
    log("FUNCTION main_display_status()");
    sprintf(log_msg, "    process_num: %i", process_num);
    log(log_msg);
    free(log_msg); log_msg = calloc(256, sizeof(char));
    sprintf(log_msg, "    process_load_status: %i", process_load_status);
    log(log_msg);
    free(log_msg); log_msg = calloc(256, sizeof(char));
    sprintf(log_msg, "    init_run: %i", init_run);
    log(log_msg);
    free(log_msg);
    log("");
    fflush(stdout);

    // Travel to desired terminal row. Skip if first loop through thread displaying.
    if (init_run == 0) {
        for (index = 0; index < (thread_args_ptr->total_processors - process_num); index ++) {
            terminal_line_up();
        }
    }
    // Clear row.
    terminal_line_erase();
    terminal_line_start();

    // Display message.
    printf("Thread %5i: %5i / %5i Loads Remaining\n", process_num, process_load_status, thread_args_ptr->total_loads);

    // Travel back to original terminal location. Skip if first loop through thread displaying.
    if (init_run == 0) {
        for (index = 0; index < (thread_args_ptr->total_processors - process_num); index ++) {
            terminal_line_down();
        }
        terminal_line_start();
    }
}


/**
 * Sends message to all worker processes that all work has been completed, and it's okay to terminate now.
 */
void main_terminate_workers(thread_struct* thread_args_ptr) {
    int program_end_tag = 1;
    int index;
    MPI_Request request_var;    // What is this for?? Why is it required?

    log("");
    log("FUNCTION main_terminate_workers()");
    log("");
    fflush(stdout);

    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        MPI_Isend(&program_end_tag, 1, MPI_INT, index, program_end_tag, MPI_COMM_WORLD, &request_var);
        // printf("Sending termination message to process %i.\n", index);
        // MPI_Send(&program_end_tag, 1, MPI_INT, index, program_end_tag, MPI_COMM_WORLD);
        // printf("Process %i received termination message.\n", index);
    }

}


/**
 * Sends worker load status and rank to master process.
 */
void worker_send_status(int process_rank, int process_load_status) {
    int* send_array = calloc(2, sizeof(int));
    char* log_msg = calloc(256, sizeof(char));

    log("");
    log("FUNCTION worker_send_status()");
    sprintf(log_msg, "    process_load_status: %i", process_load_status);
    log(log_msg);
    free(log_msg);
    log("");
    fflush(stdout);

    send_array[0] = process_rank;
    send_array[1] = process_load_status;
    MPI_Send(send_array, 2, MPI_INT, 0, 0, MPI_COMM_WORLD);

    free(send_array);
}


/**
 * Checks status of program runtime for worker to eventually get the okay to terminate.
 *
 * :return: 0 if received okay from main process to end. Otherwise 1 to continue working.
 */
int worker_check_status() {
    int msg_status_flag = 0;
    int temp = 0;
    int main_process = 0;
    int program_end_tag = 1;

    log("");
    log("FUNCTION worker_check_status()");
    log("");
    fflush(stdout);

    // Check for pending message from main process.
    // Should only ocurr if main is telling child process that all work has completed.
    MPI_Iprobe(main_process, program_end_tag, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
    if (msg_status_flag == 1) {
        // Message present. Officially grab message so master can unblock.
        MPI_Recv(&temp, 1, MPI_INT, main_process, program_end_tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        return 0;
    }
    return 1;
}


/**
 * Attempts to receive work from given "possible donor" process (donor is determined by load balancing schema).
 *
 * Does so by sending message to "possible donor" process, to initiate request.
 * Once it gets around to message checking, donor will then send back a message of:
 *  * -1 indicating donor does not have work to give.
 *  * Positive integer of work being passed back to begging process.
 */
int worker_send_request(thread_struct* thread_args_ptr, int possible_donor) {
    int donor_response = 0;
    int msg_status_flag = 0;
    char* log_msg = calloc(256, sizeof(char));

    log("");
    log("FUNCTION worker_send_request()");
    sprintf(log_msg, "    possible_donor: %i", possible_donor);
    log(log_msg);
    free(log_msg);
    log("");
    fflush(stdout);

    // Validate donor value.
    if (possible_donor >= thread_args_ptr->total_processors || possible_donor == 0) {
        // Invalid. Exit with error.
        exit(2);
    }

    // Ensure thread isn't currently sending a message to this thread. If so, abort.
    MPI_Iprobe(possible_donor, MPI_ANY_TAG, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
    if (msg_status_flag != 1) {
        // Initialilze interaction by sending a "begging message" for work.
        MPI_Send(&donor_response, 1, MPI_INT, possible_donor, 0, MPI_COMM_WORLD);

        // If got this far, then we have unblocked. Means that "possible donor" process is sending a response.
        MPI_Recv(&donor_response, 1, MPI_INT, possible_donor, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }

    // Check what response was and handle accoringly.
    if (donor_response > 0) {
        // Donor sent workload to process.
        thread_args_ptr->remaining_loads = donor_response;
        return 1;
    } else {
        // Donor sent reject message.
        return 0;
    }
}


/**
 * Checks if worker has any pending work requests from other processes.
 * On any pending request, checks if request exists on all processes.
 *
 * To handle request, either splits current work load in half (asking process gets the smaller portion if odd number)
 * or returns negative number if no work to send.
 */
void worker_handle_request(thread_struct* thread_args_ptr) {
    int msg_status_flag = 0;
    int work_send_value = 0;
    int temp = 0;
    int index;

    log("");
    log("FUNCTION worker_handle_request()");
    log("");
    fflush(stdout);

    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
    if (msg_status_flag == 1) {

        // One or more messages present. Loop through and check for message from all processors.
        for (index = 1; index < thread_args_ptr->total_processors; index++) {
            msg_status_flag = 0;
            MPI_Iprobe(index, MPI_ANY_TAG, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // Message present from processor at corresponding index. Handle.

                // First, properly "receive" pending message, so asking worker can unblock.
                MPI_Recv(&temp, 1, MPI_INT, index, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Check if donating worker has work to split and send.
                if (thread_args_ptr->remaining_loads > 1) {
                    // Worker has at least one load to split and send. Do so.
                    work_send_value = thread_args_ptr->remaining_loads / 2;
                    thread_args_ptr->remaining_loads = work_send_value + (thread_args_ptr->remaining_loads % 2);
                    MPI_Send(&work_send_value, 1, MPI_INT, index, 0, MPI_COMM_WORLD);
                } else {
                    // Worker has exactly 0 or 1 loads. Not enough to send. Reject request instead.
                    work_send_value = -1;
                    MPI_Send(&work_send_value, 1, MPI_INT, index, 0, MPI_COMM_WORLD);
                }

            }
        }

    }
}
